# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/5/26 20:12
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""

import functools
import inspect


class DelayedFunction:

    def __init__(self, func):
        self.func = func
        self._fn_params_k = inspect.signature(self.func).parameters.keys()
        self.stored_kwargs = self._get_default_args(func)
        if hasattr(func, 'stored_kwargs'):
            self.stored_kwargs.update(func.stored_kwargs)

    def _get_default_args(self, func):
        signature = inspect.signature(func)
        return {
            k: v.default
            for k, v in signature.parameters.items()
            if v.default is not inspect.Parameter.empty
        }

    def __call__(self, *args, **kwargs):
        def delayed(*args, **_kwargs):
            new_kwargs = {k: v for k, v in self.stored_kwargs.items()}
            for k, v in _kwargs.items():
                if k not in self._fn_params_k:
                    continue
                new_kwargs[k] = v
            return self.func(*args, **new_kwargs)

        self._stored_kwargs(**kwargs)
        new_fn = functools.wraps(self.func)(delayed)
        new_fn.stored_kwargs = self.stored_kwargs
        return new_fn

    def _stored_kwargs(self, **kwargs):
        for k, v in kwargs.items():
            if k not in self._fn_params_k:
                continue
            self.stored_kwargs[k] = v


def delay(func):
    """
    延迟执行
    Parameters
    ----------
    func: Callable
        需要延迟执行的对象, 必须是一个Callable对象

    Returns
    -------
    DelayedFunction
        将预先设置好的参数包装进原始的Callable对象中

    Examples
    --------

    场景1：基本使用

    >>> fn = delay(lambda a, b: a+b)(a=1, b=2)
    >>> fn()
    3

    场景2: 逐步传递参数

    >>> fn1 = delay(lambda a, b, c: a+b+c)(a=1)
    >>> fn2 = delay(fn1)(b=2)
    >>> fn2(c=3)
    6

    场景3: 参数更改

    >>> fn1 = delay(lambda a, b, c: a+b+c)(a=1, b=2)
    >>> fn2 = delay(fn1)(c=3, b=5)
    >>> fn2()
    9
    """
    return DelayedFunction(func)




# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License. 
Created on 2025/6/29 10:36
Email: yundi.xxii@outlook.com
Description:
---------------------------------------------
"""

from typing import Any
from hydra._internal.utils import _locate

class LazyImport:
    def __init__(self, full_name: str):
        self._full_name = full_name
        self._obj = None

    def _load(self) -> Any:
        if self._obj is None:
            self._obj = _locate(self._full_name)
        return self._obj

    def __getattr__(self, attr: str) -> Any:
        obj = self._load()
        return getattr(obj, attr)

    def __dir__(self) -> list[str]:
        obj = self._load()
        return dir(obj)

    def __call__(self, *args, **kwargs) -> Any:
        obj = self._load()
        if isinstance(obj, type):
            return obj(*args, **kwargs)
        elif callable(obj):
            return obj(*args, **kwargs)
        else:
            raise TypeError(f"The target `{self._full_name}` is not callable or instantiable.")

    def __repr__(self):
        return f"<LazyImport '{self._full_name}'>"

def lazy_import(full_name: str):
    """实现模块和方法的懒加载"""
    return LazyImport(full_name)


# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/5/26 23:31
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""

import inspect
import os
from functools import wraps
from pathlib import Path
import warnings
from .delay import delay

def deprecated(use_instead: str = None):
    """
    标记方法为弃用

    Parameters
    ----------
    use_instead: str
        推荐替代使用的方法或者类名称
    Returns
    -------

    """

    def decorator(func):
        """
        装饰器
        Parameters
        ----------
        func: callable
            被装饰的函数
        Returns
        -------

        """
        @wraps(func)
        def wrapper(*args, **kwargs):
            msg = f"`{func.__name__}` is deprecated. "
            if use_instead:
                msg += f"Please use `{use_instead}` instead."
            warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)
        return wrapper
    return decorator

def fn_params(func: callable):
    """
    获取fn的参数
    Parameters
    ----------
    func: callable
        需要获取参数的callable对象
    Returns
    -------
    list[tuple]

    """
    stored = delay(func)().stored_kwargs.items()
    return sorted(stored)


def fn_signature_params(func: callable):
    """获取fn所有定义的参数"""
    return sorted(list(inspect.signature(func).parameters.keys()))


def fn_path(fn: callable) -> str:
    """
    获取func所在的模块层级结构
    Parameters
    ----------
    fn: callable
        需要获取结构的callable对象
    Returns
    -------
    str
        用 `.` 连接各级层级
    """
    module = fn.__module__
    # 检查模块是否有 __file__ 属性
    if module.startswith('__main__'):
        if hasattr(module, '__file__'):
            module = module.__file__
        else:
            # 如果在交互式环境中，返回 None 或者一个默认值
            module = "<interactive environment>"
    if module.endswith('.py'):
        module = module.split('.py')[0].split(str(Path(__file__).parent.parent.absolute()))[-1]
        module = '.'.join(module.strip(os.sep).split(os.sep))
    return module


def fn_code(fn: callable) -> str:
    """
    返回fn具体的定义代码

    Parameters
    ----------
    fn: callable
        需要获取具体定义代码的callable对象

    Returns
    -------
    str
        以字符串封装定义代码

    Examples
    --------

    >>> def test_fn(a, b=2):
    >>>     return a+b
    >>> print(fn_code())
    def test_fn(a, b=2):
        return a+b
    """
    return inspect.getsource(fn)


def fn_info(fn: callable) -> str:
    """获取函数的fn_mod, params, code"""
    # mod = fn_path(fn)
    params = fn_params(fn)
    code = fn_code(fn)
    all_define_params = sorted(list(inspect.signature(fn).parameters.keys()))

    default_params = {k: v for k, v in params}
    params_infos = list()
    for p in all_define_params:
        if p in default_params:
            params_infos.append(f'{p}={default_params[p]}')
        else:
            params_infos.append(p)
    params_infos = ', '.join(params_infos)

    s = f"""
=============================================================
{fn.__name__}({params_infos})
=============================================================
{code}
    """
    return s

@deprecated("lazy_import")
def fn_from_str(s):
    """
    字符串导入对应fn
    s: a.b.c.func
    Parameters
    ----------
    s: str
        模块的路径，分隔符 `.`
    """
    import importlib
    *m_path, func = s.split(".")
    m_path = ".".join(m_path)
    mod = importlib.import_module(m_path)
    _callable = getattr(mod, func)
    return _callable

@deprecated("lazy_import")
def module_from_str(s):
    """字符串导入模块"""
    import importlib
    m_path = ".".join(s.split('.'))
    mod = importlib.import_module(m_path)
    return mod



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License. 
Created on 2025/6/29 10:49
Email: yundi.xxii@outlook.com
Description: 
---------------------------------------------
"""

import ygo

def test_import_module():
    import os
    os_path = ygo.lazy_import("os.path")
    assert os_path.join("a", "b") == os.path.join("a", "b")

def test_import_function():
    from os.path import join
    assert ygo.lazy_import("os.path.join")("a", "b") == join("a", "b")



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/4/28 15:25
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""

from .exceptions import FailTaskError, WarnException
from .pool import pool
from .delay import delay
from .utils import (
    fn_params,
    fn_signature_params,
    fn_path,
    fn_code,
    fn_info,
    module_from_str,
    fn_from_str,
)
from .lazy import lazy_import

__version__ = "v1.1.5"

__all__ = [
    "FailTaskError",
    "delay",
    "WarnException",
    "fn_params",
    "fn_signature_params",
    "fn_path",
    "fn_code",
    "fn_info",
    "fn_from_str",
    "module_from_str",
    "pool",
    "lazy_import"
]